/********************************************************************************
 * Author:  Christoph Singewald						             				*
 * develop@chenos.com								             				*
 * www.chenos.com																*
 * see FXXDatePicker.h										             		*
 *********************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif



#include "FXXDatePicker.hpp"

using namespace FX;

/*******************************************************************************/



FXDEFMAP(FXXDatePickerLabel) FXXDatePickerLabelMap[]={
  FXMAPFUNC(SEL_CLICKED,0,FXXDatePickerLabel::onClicked),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXXDatePickerLabel::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXXDatePickerLabel::onLeftBtnRelease),
  FXMAPFUNC(SEL_DOUBLECLICKED,0,FXXDatePickerLabel::onDblClicked),
  FXMAPFUNC(SEL_TRIPLECLICKED,0,FXXDatePickerLabel::onTrplClicked),
  FXMAPFUNC(SEL_PAINT,0,FXXDatePickerLabel::onPaint)
};

// Object implementation
FXIMPLEMENT(FXXDatePickerLabel,FXLabel,FXXDatePickerLabelMap,ARRAYNUMBER(FXXDatePickerLabelMap))

FXXDatePickerLabel::FXXDatePickerLabel(FXComposite* p,FXObject* tgt,FXSelector sel,const FXString& text,FXIcon* ic,FXuint opts,FXint x,FXint y,FXint w,FXint h,FXint pl,FXint pr,FXint pt,FXint pb) :
FXLabel(p,text,ic,opts,x,y,w,h,pl,pr,pt,pb){
  selector = sel;
  target = tgt;
  istoday = false;
  dayvalue=-1;

  /** DefaultColors **/
  todayColor=FXRGB(200,0,0);
}



long FXXDatePickerLabel::onLeftBtnPress(FXObject*,FXSelector,void *ptr){
  if(isEnabled()){
    grab();
    if(target &&
       target->tryHandle(this,FXSEL(SEL_LEFTBUTTONPRESS,message),ptr))
      return 1;
    return 1;
  }
  return 0;
}



long FXXDatePickerLabel::onLeftBtnRelease(FXObject*,FXSelector,void *ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;

    if(event->click_count==1){
      handle(this,FXSEL(SEL_CLICKED, 0), 0);
    }
    else if(event->click_count==2){
      handle(this,FXSEL(SEL_DOUBLECLICKED,0), 0);
    }
    else if(event->click_count==3){
      handle(this,FXSEL(SEL_TRIPLECLICKED,0), 0);
    }
    return 1;
  }

  return 0;
}



long FXXDatePickerLabel::onClicked(FXObject*,FXSelector,void*){
  if(target && target->handle(this,FXSEL(SEL_CLICKED, selector),0)){
    return 1;
  }
  return 0;
}



long FXXDatePickerLabel::onDblClicked(FXObject*,FXSelector,void*){
  if(target && target->handle(this,FXSEL(SEL_DOUBLECLICKED, selector),0)){
    return 1;
  }
  return 0;
}



long FXXDatePickerLabel::onTrplClicked(FXObject*,FXSelector,void*){
  if(target && target->handle(this,FXSEL(SEL_TRIPLECLICKED, selector),0)){
    return 1;
  }
  return 0;
}



long FXXDatePickerLabel::onPaint(FXObject* obj,FXSelector sel,void* ptr){

  FXLabel::onPaint(obj,sel,ptr);
  FXEvent   *ev=(FXEvent*)ptr;
  FXDCWindow dc(this,ev);
  if(istoday) {
    dc.setForeground(todayColor);
    dc.drawRectangle (0,0, getWidth()-1, getHeight()-1);
  }

  return 1;
}

/*******************************************************************************/


FXDEFMAP(FXXDatePicker) FXXDatePickerMap[]={
  FXMAPFUNC(SEL_CLICKED,FXXDatePicker::ID_LABEL,FXXDatePicker::onLabelClicked),
  FXMAPFUNC(SEL_DOUBLECLICKED,FXXDatePicker::ID_LABEL,FXXDatePicker::onLabelSelected),
  FXMAPFUNC(SEL_COMMAND,FXXDatePicker::ID_BUTTON_PREV,FXXDatePicker::onCmdPrev),
  FXMAPFUNC(SEL_COMMAND,FXXDatePicker::ID_BUTTON_TODAY,FXXDatePicker::onCmdToday),
  FXMAPFUNC(SEL_COMMAND,FXXDatePicker::ID_BUTTON_NEXT,FXXDatePicker::onCmdNext)
};

// Object implementation
FXIMPLEMENT(FXXDatePicker,FXVerticalFrame,FXXDatePickerMap,ARRAYNUMBER(FXXDatePickerMap))



/*******************************************************************************/


// Init
FXXDatePicker::FXXDatePicker(){
  selected=NULL;
}



FXXDatePicker::FXXDatePicker(FXComposite* p,FXXDatePickerHoliday *holi,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h,FXint pl,FXint pr,FXint pt,FXint pb):
FXVerticalFrame(p,opts,x,y,w,h,pl,pr,pt,pb)
  ,actualDate(FXDate::localDate()){

    holiday = holi;
    selector = sel;
    target =tgt;
    selected =NULL;
    /* Default Colors */
    monthbkcolor = FXRGB(184,186,222);
    yearbkcolor = FXRGB(184,186,222);
    boderColorToday = FXRGB(200,0,0);
    selectColor = FXRGB(200,200,200);
    offdaycolor = FXRGB(255,0,0);
    holidaycolor = FXRGB(0,0,255);


    FXFont * f = getApp()->getNormalFont();
    FXint textheight = f->getTextHeight ("><");
    FXint textwidth = f->getTextWidth("30"); // possible max width


    setBackColor(FXRGB(255,255,255));
    /** Creating Components **/
    navframe = new FXHorizontalFrame(this,FRAME_GROOVE|LAYOUT_FILL_X,0,0,textheight,0,0,0,0,0);
    navframe->setBackColor(yearbkcolor);
    prev= new FXArrowButton(navframe,this, ID_BUTTON_PREV,FRAME_THICK|FRAME_RAISED|ARROW_LEFT| LAYOUT_CENTER_Y,0,0,0,0,0,0,0,0);
    monthname = new FXLabel(navframe,"Januar",NULL,FRAME_NONE|LAYOUT_FILL_X);
    monthname->setBackColor(monthbkcolor);
    year = new FXLabel(navframe,"2004",NULL,FRAME_NONE);
    year->setBackColor(yearbkcolor);
    next = new FXArrowButton(navframe,this, ID_BUTTON_NEXT,FRAME_THICK|FRAME_RAISED| LAYOUT_CENTER_Y |LAYOUT_RIGHT|ARROW_RIGHT,0,0,0,0,0,0,0,0);
    FXMatrix * days =new FXMatrix(this,7,MATRIX_BY_COLUMNS| LAYOUT_FILL_COLUMN| LAYOUT_SIDE_TOP| LAYOUT_SIDE_RIGHT|LAYOUT_FILL_X,0,0,0,0,0,0,0,0);
    days->setBackColor(FXRGB(255,255,255));


    FXLabel * l;
    l = new FXLabel(days,FXDate::dayNameShort(1),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l = new FXLabel(days,FXDate::dayNameShort(2),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l = new FXLabel(days,FXDate::dayNameShort(3),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l = new FXLabel(days,FXDate::dayNameShort(4),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l = new FXLabel(days,FXDate::dayNameShort(5),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l = new FXLabel(days,FXDate::dayNameShort(6),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l->setTextColor(offdaycolor);
    l = new FXLabel(days,FXDate::dayNameShort(0),NULL,FRAME_NONE);
    l->setBackColor(FXRGB(255,255,255));
    l->setTextColor(offdaycolor);

    FXint count  = 0;
    for(FXint i=0; i<42;i++)
      {
        FXXDatePickerLabel *l = new FXXDatePickerLabel(days,this,ID_LABEL," ",NULL,FRAME_NONE | JUSTIFY_CENTER_X | JUSTIFY_CENTER_Y | LAYOUT_FILL_X| LAYOUT_FILL_Y |LAYOUT_CENTER_X | LAYOUT_CENTER_Y,0,0,textwidth,0);
        l->setBackColor(FXRGB(255,255,255));
        dayfields.append(l);
        if(count == 5 || count==6) {
          l->setTextColor(offdaycolor);
        }
        if(count==6) count = 0; else  count++;

      }

    selectedDate=actualDate;
    fill(actualDate);

  }

FXXDatePicker::~FXXDatePicker()
  {
    if(holiday) delete holiday;
  }



// Create window
void FXXDatePicker::create(){
  FXVerticalFrame::create();
}


// Detach window
void FXXDatePicker::detach(){
  FXVerticalFrame::detach();

}
void FXXDatePicker::setMonthBkColor(FXColor col)
{
  monthbkcolor=col;
  navframe->setBackColor(monthbkcolor);
  monthname->setBackColor(monthbkcolor);
}
void FXXDatePicker::setYearBkColor(FXColor col)
{
  yearbkcolor=col;
  navframe->setBackColor(yearbkcolor);
  year->setBackColor(yearbkcolor);
}
void FXXDatePicker::setSelectionColor(FXColor col)
{
  selectColor=col;
}
void FXXDatePicker::setOffdayColor(FXColor col)
{
  offdaycolor=col;
  FXint count = 0;
  for(FXint i=0; i<42;i++)
    {
      if(count == 5 || count==6) {
        ((FXLabel * )dayfields[i])->setTextColor(offdaycolor);
      }
      if(count==6) count = 0; else  count++;

    }
}
void FXXDatePicker::setHolidayColor(FXColor col)
{
  holidaycolor=col;
}



long FXXDatePicker::onLabelClicked(FXObject* obj,FXSelector sel,void* ptr){
  if(((FXXDatePickerLabel*)obj)->getDayValue()==-1)
    return 1;

  if (selected) {
    selected->setBackColor(getBackColor());
  }
  selected = (FXXDatePickerLabel *)obj;
  selected->setBackColor(selectColor);
  // We have a date selected send a message to target
  selectedDate.setDate(actualDate.year(),
                       actualDate.month(),
                       selected->getDayValue());
  if(target)
    target->handle(this,FXSEL(SEL_CHANGED,selector),&selectedDate);

  return 1;
}



long FXXDatePicker::onLabelSelected(FXObject* obj,FXSelector sel,void* ptr){
  if(((FXXDatePickerLabel*)obj)->getDayValue()==-1)
    return 1;

  if(selected) {
    selected->setBackColor(getBackColor());
  }
  selected = (FXXDatePickerLabel *)obj;
  selected->setBackColor(selectColor);
  // We have a date selected send a message to target
  selectedDate.setDate(actualDate.year(),
                       actualDate.month(),
                       selected->getDayValue());
  if(target)
    target->handle(this,FXSEL(SEL_SELECTED,selector),&selectedDate);
  return 1;
}



long FXXDatePicker::onCmdPrev(FXObject* obj,FXSelector sel,void* ptr){
  int y, m, d;
  y=actualDate.year();
  m=actualDate.month()-1;
  d=actualDate.day();
  if (m<1) {
    m=12;
    y--;
  }
  actualDate.setDate(y, m, d);
  fill(actualDate);
  if(target)
    target->handle(this,FXSEL(SEL_CHANGED,selector),&actualDate);
  return 0;
}



long FXXDatePicker::onCmdToday(FXObject* obj,FXSelector sel,void* ptr){

  actualDate = FXDate::localDate();
  fill(actualDate);
  if(target) target->handle(this,FXSEL(SEL_CHANGED,selector),&actualDate);

  return 0;
}
long FXXDatePicker::onCmdNext(FXObject* obj,FXSelector sel,void* ptr)
{
  int y, m, d;
  y=actualDate.year();
  m=actualDate.month()+1;
  d=actualDate.day();
  if (m>12) {
    m=1;
    y++;
  }
  actualDate.setDate(y, m, d);
  fill(actualDate);
  if(target) target->handle(this,FXSEL(SEL_CHANGED,selector),&actualDate);
  return 0;
}

void FXXDatePicker::reset()
{
  actualDate = FXDate::localDate();
  fill(actualDate);

}


void FXXDatePicker::setDay(FXDate &date, FXint day) {
  FXint y, m;

  y=date.year();
  m=date.month();
  date.setDate(y, m, day);
}


void FXXDatePicker::fill(FXDate &adate)
{

  FXint i = 0;
  FXString buffer ="";
  FXDate firstdayofmonth(adate.year(),adate.month(),1);
  FXDate today=FXDate::localDate();
  FXDate tempdate =firstdayofmonth;
  if (selected){
    selected->setBackColor(getBackColor());
  }
  selected=NULL;

  monthname->setText(adate.monthName(adate.month()));
  buffer.format("%04d",adate.year());
  year->setText(buffer);

  FXint dayindex = (firstdayofmonth.dayOfWeek()==0)?6:firstdayofmonth.dayOfWeek()-1;

  // Delete unused

  for(i=0; i<dayfields.no();i++) {
    ((FXLabel * )dayfields[i])->setText(" ");
    ((FXXDatePickerLabel * )dayfields[i])->setToday(false);
    ((FXXDatePickerLabel * )dayfields[i])->setDayValue(-1);
  }


  FXDate d=adate;
  for(i=dayindex; i<adate.daysInMonth()+dayindex;i++) {
    buffer.format("%d",i-dayindex+1);
    ((FXLabel * )dayfields[i])->setText(buffer);
    ((FXXDatePickerLabel * )dayfields[i])->setDayValue(i-dayindex+1);

    setDay(d, i-dayindex+1);
    if (selectedDate == d){
      selected = (FXXDatePickerLabel * )dayfields[i];
      selected->setBackColor(selectColor);
    }
    if(today == d) {
      ((FXXDatePickerLabel * )dayfields[i])->setToday();
      ((FXXDatePickerLabel * )dayfields[i])->setTodayColor(boderColorToday);
    }
    if(holiday) {
      setDay(tempdate, i-dayindex+1);
      if(holiday->isHoliday(tempdate))
        {
          ((FXLabel * )dayfields[i])->setTextColor(holidaycolor);
        }
      else if(tempdate.dayOfWeek()==0 || tempdate.dayOfWeek()==6) {
        ((FXLabel * )dayfields[i])->setTextColor(offdaycolor);
      }
      else
        ((FXLabel * )dayfields[i])->setTextColor(FXRGB(0,0,0));
    }
  }
}

FXDate  FXXDatePicker::getSelectedDate()
{
  /*if(selected)
   * {
   actualDate.setDay(selected->getDayValue());
   }
   return actualDate;*/
  return selectedDate;

}

FXDate  FXXDatePicker::setSelectedDate(FXDate d)
{
  actualDate=d;
  selectedDate=d;
  fill(actualDate);

  return actualDate;
}


