/***************************************************************************
 * This file is part of the project AqFoxExt.                              *
 * AqFoxExt (c) by 2011 Martin Preuss, all rights reserved.                *
 *                                                                         *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "fxx_texttable.hpp"


#include <gwenhywfar/debug.h>

#include <assert.h>



FXXTextTable::FXXTextTable(FXComposite *p, FXObject* tgt, FXSelector sel, FXuint opts,
                           FXint x, FXint y, FXint w, FXint h,
                           FXint pl, FXint pr,
                           FXint pt, FXint pb)
:FXXTable(p, tgt, sel, opts, x, y, w, h, pl, pr, pt, pb)
,m_strings(NULL) {

}



FXXTextTable::FXXTextTable()
:FXXTable()
,m_strings(NULL) {
}



FXXTextTable::~FXXTextTable() {
  if (m_strings) {
    delete [] m_strings;
    m_strings=NULL;
  }
}



void FXXTextTable::setNumberOfRows(int i) {
  if (i!=getNumberOfRows()) {
    if (m_strings) {
      FXString *p;
      int l;
      int cnt;
      int j;

      l=getNumberOfColumns()*i;
      p=new FXString[l];
      assert(p);

      cnt=i;
      if (i>getNumberOfRows())
        cnt=getNumberOfRows();
      cnt*=getNumberOfColumns();
      for (j=0; j<cnt; j++)
        p[j]=m_strings[j];
      delete [] m_strings;
      m_strings=p;
    }

    FXXTable::setNumberOfRows(i);
  }
}



void FXXTextTable::addColumn(const FXString &s, int size) {
  if (m_strings) {
    FXString *p;
    int r;

    p=new FXString[(getNumberOfColumns()+1)*getNumberOfRows()];
    assert(p);

    for (r=0; r<getNumberOfRows(); r++) {
      int j;
      int srcpos=getNumberOfColumns()*r;
      int dstpos=(getNumberOfColumns()+1)*r;

      for (j=0; j<getNumberOfColumns(); j++) {
        p[dstpos+j]=m_strings[srcpos+j];
      }
    }

    delete [] m_strings;
    m_strings=p;
  }

  FXXTable::addColumn(s, size);
}



const FXString &FXXTextTable::getText(int r, int c) {
  if (r>=0 && c>=0 && r<getNumberOfRows() && c<getNumberOfColumns() && m_strings)
    return m_strings[(r*getNumberOfColumns())+c];
  return m_emptyString;
}



void FXXTextTable::setText(int r, int c, const FXString &s) {
  if (r>=0 && c>=0 && r<getNumberOfRows() && c<getNumberOfColumns()) {
    if (m_strings==NULL) {
      DBG_ERROR(0, "Creating %d strings", getNumberOfColumns()*getNumberOfRows());
      m_strings=new FXString[getNumberOfColumns()*getNumberOfRows()];
    }
    m_strings[(r*getNumberOfColumns())+c]=s;
    update();
  }
}



FXXTable::ColumnInfo *FXXTextTable::getColumnInfo(int r, int c) {
  return new ColumnInfoText(this, r, c, getText(r, c));
}



void FXXTextTable::releaseColumnInfo(ColumnInfo *ci) {
  delete ci;
}



FXWindow *FXXTextTable::getEditor(int r, int c) {
  FXTextField *ed=new FXTextField(this, 1, NULL, 0,
                                  TEXTFIELD_ENTER_ONLY,
                                  0, 0, 0, 0,
                                  0, 0, 0, 0);
  ed->create();
  ed->setFont(m_font);
  ed->setBackColor(getCellBackColor(r, c));
  ed->setTextColor(m_textColor);
  ed->setSelBackColor(m_selbackColor);
  ed->setSelTextColor(m_seltextColor);
  ed->setText(getText(r, c));
  ed->selectAll();

  return ed;
}



void FXXTextTable::setFromEditor(ColumnInfo *ci, FXWindow *ed) {
  DBG_ERROR(0, "SetFromEditor");
  FXTextField *tf=dynamic_cast<FXTextField*>(ed);

  if (tf) {
    ColumnInfoText *tci=dynamic_cast<ColumnInfoText*>(ci);

    if (tci) {
      setText(tci->getRow(), tci->getColumn(), tf->getText());
      tci->setText(tf->getText());
      ci->update();
    }
  }
  else {
    DBG_ERROR(0, "Not a text field");
  }
}



void FXXTextTable::releaseEditor(int r, int c, FXWindow *ed) {
  delete ed;
}






