/***************************************************************************
 * This file is part of the project AqFoxExt.                              *
 * AqFoxExt (c) by 2011 Martin Preuss, all rights reserved.                *
 *                                                                         *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "fxx_tooltip.hpp"

#include <gwen-gui-fox16/fox16_htmlctx.hpp>



FXDEFMAP(FXXToolTip) FXXToolTipMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXXToolTip::onPaint),
  FXMAPFUNC(SEL_COMMAND,FXToolTip::ID_SETSTRINGVALUE, FXXToolTip::onCmdSetStringValue),
};


// Object implementation
FXIMPLEMENT(FXXToolTip, FXToolTip,FXXToolTipMap, ARRAYNUMBER(FXXToolTipMap))





static const char *_strCaseStr(const char *haystack, const char *needle) {
  while(*haystack) {
    while(*haystack && tolower(*haystack)!=tolower(*needle))
      haystack++;
    if (*haystack) {
      const char *t;
      const char *s;

      /* now haystack points to an area which begins with *needle */
      t=haystack;
      s=needle;
      t++;
      s++;
      while(*t && *s && (tolower(*t)==tolower(*s))) {
	t++;
	s++;
      }
      if (*s==0)
	return haystack;
    }
    else
      /* not found */
      break;
    /* no match here, advance haystack */
    haystack++;
  }

  /* not found */
  return NULL;
}



FXXToolTip::FXXToolTip(FXApp* a, FXuint opts, FXint x, FXint y, FXint w, FXint h)
:FXToolTip(a, opts, x, y, w, h)
,m_isHtml(false)
,m_htmlCtx(NULL)
,m_haveTextDims(false)
,m_textWidth(0)
,m_textHeight(0)
{
}



FXXToolTip::FXXToolTip()
:FXToolTip()
,m_isHtml(false)
,m_htmlCtx(NULL)
,m_haveTextDims(false)
,m_textWidth(0)
,m_textHeight(0)
{
}



FXXToolTip::~FXXToolTip() {
}



void FXXToolTip::calcTextDims() {
  int wNeeded;
  int w;

  if (m_htmlCtx==NULL)
    updateHtml();

  m_htmlCtx->layout(-1, -1);
  wNeeded=m_htmlCtx->getWidth();
  w=wNeeded;
  if (w<width)
    w=width;
  m_textWidth=w+2;
  m_textHeight=m_htmlCtx->getHeight()+2;
  m_haveTextDims=true;
}



void FXXToolTip::updateHtml() {
  if (m_htmlCtx)
    delete m_htmlCtx;
  m_htmlCtx=new FOX16_HtmlCtx(0);
  m_htmlCtx->setForegroundColor(textColor);
  m_htmlCtx->setBackgroundColor(backColor);
  m_htmlCtx->setText(label.text());

  flags|=FLAG_DIRTY;
}



FXint FXXToolTip::getDefaultWidth() {
  if (!m_isHtml)
    return FXToolTip::getDefaultWidth();
  else {
    if (!m_haveTextDims)
      calcTextDims();
    return m_textWidth;
  }
}



FXint FXXToolTip::getDefaultHeight() {
  if (!m_isHtml)
    return FXToolTip::getDefaultHeight();
  else {
    if (!m_haveTextDims)
      calcTextDims();
    return m_textHeight;
  }
}



void FXXToolTip::layout() {
  if (!m_isHtml)
    return FXToolTip::layout();
  else {
    m_haveTextDims=false;
    if (m_htmlCtx==NULL)
      updateHtml();
    m_htmlCtx->layout(width, height);
    update();
    flags&=~FLAG_DIRTY;
  }
}



long FXXToolTip::onCmdSetStringValue(FXObject*, FXSelector, void *ptr) {
  const FXString *str=(FXString*)ptr;

  if (str && _strCaseStr(str->text(), "<html>")) {
    m_isHtml=true;
    m_haveTextDims=false;
    FXToolTip::setText(*str);
    updateHtml();

    flags|=FLAG_DIRTY;
    layout();
    recalc();
    update();

    return 1;
  }
  else {
    m_isHtml=false;
    FXToolTip::setText(*str);
    return 1;
  }
}



long FXXToolTip::onPaint(FXObject *sender, FXSelector sel, void *ptr) {
  if (!m_isHtml)
    return FXToolTip::onPaint(sender, sel, ptr);
  else {
    FXEvent *ev=(FXEvent*)ptr;
    FXDCWindow dc(this, ev);

    dc.setForeground(backColor);
    dc.fillRectangle(0, 0, width, height);

    dc.setForeground(textColor);
    dc.drawRectangle(0, 0, width-1, height-1);

    if (m_htmlCtx)
      m_htmlCtx->paint(&dc, 1, 1);

    return 1;
  }
}




